from dataclasses import dataclass
from typing import Optional, Tuple

import numpy
import torch

import torch_gauge.o3
from torch_gauge.o3.functional import NormContraction1d, NormContraction2d


@dataclass
class SphericalTensor:
    """
    The SphericalTensor class tracks the SO(3) representation indices
    of a flattened data tensor for efficient group algebra operations.

    All angular and magnetic quantum number indices are treated as equivalent,
    and all non-degenerate feature channels are regarded as the "principal"
    indices, n. Note that the angular quantum number indices (m) for each l
    range from -l to l as in the physicist's convention.

    Two indexing tensors are maintained, ``metadata`` and ``rep_layout``.

    Attributes:
        data_ten (torch.Tensor): The tensor storing auto-differentiable data.
        metadata (torch.LongTensor): The minimal information which specifies
            the number of non-degenerate feature channels for each dimension
            and each l (and the associated m(s)).
        rep_layout (Tuple[torch.LongTensor]): Stores the pointer (l[i], m[i], n[i])
            for each index i along the SO(3) representation dimension of the
            flattened data tensor.

    Args:
        data_ten (torch.Tensor): the underlying flattened data tensor in the
             strictly incremental [l,m,n] order, for which the rep_dim has N elements
        rep_dims (tuple of ints): the spherical tensor representation dimension(s). When
             n_rep_dims > 1, must be contiguous dimensions.
        metadata (torch.LongTensor): (n_rep_dims x n_l) specification of the number of unique
             channels (n) for each angular momentum index l.
        rep_layout (tuple of ints, optional): the (3xN) index tensor for (l[i],m[i],n[i]).
        num_channels (tuple of ints, optional): the number of unique channels per rep_dim.
    """

    def __init__(
        self,
        data_ten: torch.Tensor,
        rep_dims: Tuple[int, ...],
        metadata: torch.LongTensor,
        rep_layout: Optional[Tuple[torch.LongTensor, ...]] = None,
        num_channels: Optional[Tuple[int, ...]] = None,
    ):
        self.ten = data_ten
        self.metadata = metadata
        self.rep_dims = rep_dims
        assert self.rep_dims[-1] < self.ten.dim()
        self._norm_eps = 1e-4
        if rep_layout:
            self.rep_layout = rep_layout
        else:
            if len(rep_dims) > 1:
                assert rep_dims[1] - rep_dims[0] == 1
            self.rep_layout = self.generate_rep_layout()
        if num_channels:
            self.num_channels = num_channels
        else:
            self.num_channels = tuple(torch.sum(self.metadata, dim=1).long().tolist())

    @property
    def shape(self):
        return self.ten.shape

    @property
    def device(self):
        return self.ten.device

    def self_like(self, new_data_ten: torch.Tensor):
        """
        Args:
            new_data_ten: The underlying data of the new SphericalTensor.

        Returns:
             A SphericalTensor with identical layout but new data tensor.
        """
        return SphericalTensor(
            new_data_ten,
            rep_dims=self.rep_dims,
            metadata=self.metadata,
            rep_layout=self.rep_layout,
            num_channels=self.num_channels,
        )

    def mul_(self, other: "SphericalTensor"):
        # Be careful that this operation will break equivariance
        self.ten.mul_(other.ten)
        return self.ten

    def add_(self, other: "SphericalTensor"):
        self.ten.add_(other.ten)
        return self.ten

    def __mul__(self, other: "SphericalTensor"):
        return self.self_like(self.ten * other.ten)

    def __add__(self, other: "SphericalTensor"):
        return self.self_like(self.ten + other.ten)

    def __sub__(self, other: "SphericalTensor"):
        return self.self_like(self.ten - other.ten)

    def scalar_mul(self, other: torch.Tensor, inplace=False):
        """
        Performing representation-wise scalar multiplication.

        Args:
            other (torch.Tensor): along rep_dims, the shape must be the number
                of unique channels (n); along all other dimensions the shape must
                be the same as the data tensor.
            inplace (bool): If true, self.ten is updated in-place.
        """
        if len(self.rep_dims) == 1:
            broadcasted_other = torch.index_select(
                other, dim=self.rep_dims[0], index=self.rep_layout[0][2, :]
            )
        elif len(self.rep_dims) == 2:
            broadcasted_other = torch.index_select(
                other.view(
                    *other.shape[: self.rep_dims[0]],
                    self.num_channels[0] * self.num_channels[1],
                    *other.shape[self.rep_dims[1] + 1 :],
                ),
                dim=self.rep_dims[0],
                index=(
                    self.rep_layout[0][2, :].unsqueeze(1) * self.num_channels[1]
                    + self.rep_layout[1][2, :].unsqueeze(0)
                ).view(-1),
            ).view_as(self.ten)
        else:
            raise NotImplementedError

        if inplace:
            self.ten.mul_(broadcasted_other)
            return self
        else:
            return self.self_like(self.ten * broadcasted_other)

    def dot(self, other: "SphericalTensor", dim: int):
        """
        Performing inner product along a representation dimension.

        If self.n_rep_dim==1, a torch.Tensor is returned;
        if self.n_rep_dim==2, a SphericalTensor with n_rep_dim==1 is returned.

        Args:
            other (SphericalTensor): Must be a 1-d spherical tensor
                with the same number of dimensions and broadcastable to self.ten.
                When self.n_rep_dim==2, the kept dimension in self.rep_dims
                of other must be singleton.
            dim (int): The dimension to perform the inner product.
        """
        dotdim_idx = self.rep_dims.index(dim)
        assert other.rep_dims[0] == dim
        assert torch.all(self.metadata[dotdim_idx].eq(other.metadata[0]))
        out_ten = self.ten.mul(other.ten).sum(dim)
        if len(self.rep_dims) == 1:
            return out_ten
        elif len(self.rep_dims) == 2:
            dimid_kept = 1 - dotdim_idx
            assert other.ten.shape[self.rep_dims[dimid_kept]] == 1
            return self.__class__(
                out_ten,
                rep_dims=(self.rep_dims[dimid_kept],),
                metadata=self.metadata[dimid_kept].unsqueeze(0),
                rep_layout=(self.rep_layout[dimid_kept],),
                num_channels=(self.num_channels[dimid_kept],),
            )
        else:
            raise NotImplementedError

    def rep_dot(self, other: "SphericalTensor", dim: int):
        """
        Performing channel-wise inner products.

        If self.n_rep_dim==1, a torch.Tensor is returned;
        if self.n_rep_dim==2, a SphericalTensor with n_rep_dim==1 is returned.

        Args:
            other (SphericalTensor): must be a 1-d spherical tensor
                with the same number of dimensions and broadcastable to self.ten.
            dim (int): the dimension to perform the channel-wise inner product.
        """
        dotdim_idx = self.rep_dims.index(dim)
        assert other.rep_dims[0] == dim
        assert torch.all(self.metadata[dotdim_idx].eq(other.metadata[0]))
        mul_ten = self.ten * other.ten
        out_shape = list(mul_ten.shape)
        out_shape[dim] = self.num_channels[dotdim_idx]
        out_ten = torch.zeros(
            out_shape, device=mul_ten.device, dtype=mul_ten.dtype
        ).index_add_(
            dim,
            self.rep_layout[dotdim_idx][2, :],
            mul_ten,
        )
        # Note that the argument names of index_add in pytorch is changing from tensor to source
        # in 1.7-1.8. Not using keywords for compatibility
        if len(self.rep_dims) == 1:
            return out_ten
        elif len(self.rep_dims) == 2:
            # A little trick here
            dimid_kept = 1 - dotdim_idx
            return self.__class__(
                out_ten,
                rep_dims=(self.rep_dims[dimid_kept],),
                metadata=self.metadata[dimid_kept].unsqueeze(0),
                rep_layout=(self.rep_layout[dimid_kept],),
                num_channels=(self.num_channels[dimid_kept],),
            )
        else:
            raise NotImplementedError

    def rep_outer(self, other: "SphericalTensor") -> "SphericalTensor":
        """
        Returns the batched outer product of two 1-d spherical tensors.

        The rep_dim and metadata of self and other much be the same.
        """
        assert len(self.rep_dims) == 1
        assert len(other.rep_dims) == 1
        assert (
            self.rep_dims[0] == other.rep_dims[0]
        ), "The representation dimensions of self and other must be contiguous"
        odim = self.rep_dims[0]
        out_ten = self.ten.unsqueeze(odim + 1).mul(other.ten.unsqueeze(odim))
        out_metadata = torch.cat([self.metadata, other.metadata], dim=0)
        out_rep_layout = (
            self.rep_layout[0],
            other.rep_layout[0],
        )
        return self.__class__(
            out_ten,
            rep_dims=(odim, odim + 1),
            metadata=out_metadata,
            rep_layout=out_rep_layout,
            num_channels=(self.num_channels, other.num_channels),
        )

    def fold(self, stride: int, update_self=False) -> "SphericalTensor":
        """
        Fold the chucked representation channels of a 1-d SphericalTensor to a new dimension.
        """
        assert len(self.rep_dims) == 1
        assert torch.all(torch.fmod(self.metadata[0], stride) == 0), (
            f"The number of channels for the SphericalTensor to be folded must be multiples of "
            f"stride, got ({self.metadata}, {stride}) instead"
        )
        new_ten = self.ten.unflatten(
            dim=self.rep_dims[0],
            sizes=(self.shape[self.rep_dims[0]] // stride, stride),
        )
        new_metadata = self.metadata // stride
        new_rep_layout = (self.rep_layout[0][:, ::stride],)
        new_num_channels = (self.num_channels[0] // stride,)
        if update_self:
            self.ten = new_ten
            self.metadata = new_metadata
            self.rep_layout = new_rep_layout
            self.num_channels = new_num_channels
            return self
        else:
            return self.__class__(
                new_ten,
                rep_dims=self.rep_dims,
                metadata=new_metadata,
                rep_layout=new_rep_layout,
                num_channels=new_num_channels,
            )

    def unfold(self, update_self=False) -> "SphericalTensor":
        """
        Contract one trailing dimension into the representation dimension.
        """
        assert len(self.rep_dims) == 1
        assert self.ten.dim() > self.rep_dims[0], "No trailing dimension to unfold"
        stride = self.ten.shape[self.rep_dims[0] + 1]
        new_ten = self.ten.flatten(
            self.rep_dims[0],
            self.rep_dims[0] + 1,
        )
        new_metadata = self.metadata * stride
        new_rep_layout = (self.rep_layout[0].repeat_interleave(stride, dim=1),)
        new_num_channels = (self.num_channels[0] * stride,)
        if update_self:
            self.ten = new_ten
            self.metadata = new_metadata
            self.rep_layout = new_rep_layout
            self.num_channels = new_num_channels
            return self
        else:
            return self.__class__(
                new_ten,
                rep_dims=self.rep_dims,
                metadata=new_metadata,
                rep_layout=new_rep_layout,
                num_channels=new_num_channels,
            )

    def transpose_repdims(self, inplace=False):
        assert (
            len(self.rep_dims) == 2
        ), "transpose_repdims only supports 2d SphericalTensor"
        ten_t = torch.transpose(self.ten, *self.rep_dims).contiguous()
        dims_t = self.rep_dims
        metadata_t = self.metadata[(1, 0), :]
        rep_layout_t = self.rep_layout[::-1]
        num_channels_t = self.num_channels[::-1]
        if inplace:
            self.ten = ten_t
            self.rep_dims = dims_t
            self.metadata = metadata_t
            self.rep_layout = rep_layout_t
            self.num_channels = num_channels_t
            return self
        else:
            return self.__class__(
                ten_t,
                rep_dims=dims_t,
                metadata=metadata_t,
                rep_layout=rep_layout_t,
                num_channels=num_channels_t,
            )

    def invariant(self) -> torch.Tensor:
        """
        Returns the invariant content of a SphericalTensor

        When self.n_rep_dim==1, the l=0 channels are retained;
        when self.n_rep_dim==2, the (l1=0, l2=0) channels are also contracted.
        """
        if len(self.rep_dims) == 1:
            l0_length = self.metadata[0, 0]
            ops_dim = self.rep_dims[0]
            data_l0 = torch.narrow(self.ten, dim=ops_dim, start=0, length=l0_length)
            norm_shape = list(self.shape)
            norm_shape[ops_dim] = self.num_channels[0] - l0_length
            data_rep = torch.narrow(
                self.ten,
                dim=ops_dim,
                start=l0_length,
                length=self.ten.shape[ops_dim] - l0_length,
            )
            # Subtract the L=0 offset in the pointer tensor
            idx_ten = self.rep_layout[0][2, l0_length:].sub(l0_length)
            invariant_rep = NormContraction1d.apply(
                data_rep, idx_ten, norm_shape, ops_dim, self._norm_eps
            )
            return torch.cat([data_l0, invariant_rep], dim=ops_dim)
        elif len(self.rep_dims) == 2:
            idx_ten_0 = (
                self.rep_layout[0][2, :]
                .unsqueeze(1)
                .expand(
                    self.ten.shape[self.rep_dims[0]], self.ten.shape[self.rep_dims[1]]
                )
            )
            idx_ten_1 = (
                self.rep_layout[1][2, :]
                .unsqueeze(0)
                .expand(
                    self.ten.shape[self.rep_dims[0]], self.ten.shape[self.rep_dims[1]]
                )
            )
            idx_tens = torch.stack([idx_ten_0, idx_ten_1], dim=0)
            norm_shape = list(self.shape)
            norm_shape[self.rep_dims[0]] = self.num_channels[0]
            norm_shape[self.rep_dims[1]] = self.num_channels[1]
            invariant2d = NormContraction2d.apply(
                self.ten, idx_tens, norm_shape, self.rep_dims, self._norm_eps
            )
            return invariant2d
        else:
            raise NotImplementedError

    def generate_rep_layout(self) -> Tuple[torch.LongTensor, ...]:
        if len(self.rep_dims) == 1:
            return (self.generate_rep_layout_1d_(self.metadata[0]).to(self.ten.device),)
        elif len(self.rep_dims) == 2:
            rep_layout_0 = self.generate_rep_layout_1d_(self.metadata[0]).to(
                self.ten.device
            )
            rep_layout_1 = self.generate_rep_layout_1d_(self.metadata[1]).to(
                self.ten.device
            )
            return rep_layout_0, rep_layout_1
        else:
            raise NotImplementedError

    @staticmethod
    def generate_rep_layout_1d_(metadata1d) -> torch.LongTensor:
        n_irreps_per_l = torch.arange(start=0, end=metadata1d.size(0)) * 2 + 1
        end_channelids = torch.cumsum(metadata1d, dim=0)
        start_channelids = torch.cat([torch.LongTensor([0]), end_channelids[:-1]])
        dst_ls = torch.repeat_interleave(
            torch.arange(n_irreps_per_l.size(0)), n_irreps_per_l * metadata1d
        )
        dst_ms = torch.repeat_interleave(
            torch.cat(
                [
                    torch.arange(-l, l + 1)
                    for l in torch.arange(start=0, end=metadata1d.size(0))
                ]
            ),
            torch.repeat_interleave(metadata1d, n_irreps_per_l),
        )
        ns = torch.arange(metadata1d.sum())
        dst_ns = torch.cat(
            [
                ns[start_channelids[l] : end_channelids[l]].repeat(n_irreps)
                for l, n_irreps in enumerate(n_irreps_per_l)
            ]
        )
        rep_layout = torch.stack([dst_ls, dst_ms, dst_ns], dim=0).long()
        assert rep_layout.shape[1] == torch.sum(n_irreps_per_l * metadata1d)

        return rep_layout

    def to(self, device):
        # Do not transfer metadata to GPU, as index tracking should be more efficient
        # on CPUs
        self.ten = self.ten.to(device)
        self.rep_layout = tuple(layout.to(device) for layout in self.rep_layout)
        return self


def to_numpy(src_ten):
    # Translate to a databasing-friendly dictionary object
    if isinstance(src_ten, torch.Tensor):
        return {
            "_type": "torch.Tensor",
            "ten": src_ten.numpy(),
        }
    elif isinstance(src_ten, SphericalTensor):
        return {
            "_type": src_ten.__class__.__name__,
            "ten": src_ten.ten.numpy(),
            "rep_dims": numpy.asarray(src_ten.rep_dims),
            "metadata": src_ten.metadata.numpy(),
            "cated_rep_layout": torch.cat(src_ten.rep_layout, dim=1).numpy(),
            "rep_offsets": numpy.cumsum(
                numpy.asarray(list(rl.shape[1] for rl in src_ten.rep_layout))
            ),
            "num_channels": numpy.asarray(src_ten.num_channels),
        }
    elif src_ten is None:
        return {
            "_type": "NoneType",
        }
    else:
        raise ValueError


def from_numpy(src_dict):
    if src_dict["_type"] == "torch.Tensor":
        return torch.from_numpy(src_dict["ten"])
    elif src_dict["_type"] == "NoneType":
        return None
    else:
        target_class = getattr(torch_gauge.o3.spherical, src_dict["_type"])
        rep_layout = numpy.split(
            src_dict["cated_rep_layout"], src_dict["rep_offsets"][:-1], axis=1
        )
        rep_layout = tuple(torch.from_numpy(rl) for rl in rep_layout)
        return target_class(
            torch.from_numpy(src_dict["ten"]),
            rep_dims=tuple(src_dict["rep_dims"]),
            metadata=torch.from_numpy(src_dict["metadata"]),
            rep_layout=rep_layout,
            num_channels=tuple(src_dict["num_channels"]),
        )


@dataclass
class O3Tensor(SphericalTensor):
    """
    The O3Tensor class tracks the O(3) representation indices with additional
    parity (p) index tracker.

    Two indexing tensors are maintained, ``metadata`` and ``rep_layout``.

    Attributes:
        data_ten (torch.Tensor): The tensor storing auto-differentiable data.
        metadata (torch.LongTensor): The minimal information which specifies
            the number of non-degenerate feature channels for each dimension
            and each l and p (and the associated m(s)).
        rep_layout (Tuple[torch.LongTensor]): Stores the pointer (l[i], m[i], n[i], p[i])
            for each index i along the SO(3) representation dimension of the
            flattened data tensor.

    Args:
        data_ten (torch.Tensor): the underlying flattened data tensor in the
             strictly incremental [l,m,-p,n] order, for which the rep_dim has N elements
        rep_dims (tuple of ints): the spherical tensor representation dimension(s). When
             n_rep_dims > 1, must be contiguous dimensions.
        metadata (torch.LongTensor): (n_rep_dims x (2 * n_l)) specification of the number of unique
             channels (n) for each angular momentum index l.
        rep_layout (tuple of ints, optional): the (4xN) index tensor for (l[i],m[i],n[i],p[i]).
        num_channels (tuple of ints, optional): the number of unique channels per rep_dim.
    """

    def __init__(
        self,
        data_ten: torch.Tensor,
        rep_dims: Tuple[int, ...],
        metadata: torch.LongTensor,
        rep_layout: Optional[Tuple[torch.LongTensor, ...]] = None,
        num_channels: Optional[Tuple[int, ...]] = None,
    ):
        self.ten = data_ten
        self.metadata = metadata
        assert self.metadata.shape[1] % 2 == 0
        self.rep_dims = rep_dims
        assert self.rep_dims[-1] < self.ten.dim()
        self._norm_eps = 1e-4
        if rep_layout:
            self.rep_layout = rep_layout
        else:
            if len(rep_dims) > 1:
                assert rep_dims[1] - rep_dims[0] == 1
            self.rep_layout = self.generate_rep_layout()
        if num_channels:
            self.num_channels = num_channels
        else:
            self.num_channels = tuple(torch.sum(self.metadata, dim=1).long().tolist())

    def self_like(self, new_data_ten: torch.Tensor):
        """
        Args:
            new_data_ten: The underlying data of the new O3Tensor.

        Returns:
             A O3Tensor with identical layout but new data tensor.
        """
        return O3Tensor(
            new_data_ten,
            rep_dims=self.rep_dims,
            metadata=self.metadata,
            rep_layout=self.rep_layout,
            num_channels=self.num_channels,
        )

    @staticmethod
    def generate_rep_layout_1d_(metadata1d) -> torch.LongTensor:
        num_distinct_ls = metadata1d.size(0) // 2
        n_irreps_per_l = torch.arange(start=0, end=num_distinct_ls) * 2 + 1
        n_irreps_per_lp = n_irreps_per_l.repeat_interleave(2)
        ls_metadata = torch.arange(start=0, end=num_distinct_ls).repeat_interleave(2)
        end_channelids = torch.cumsum(metadata1d, dim=0)
        start_channelids = torch.cat([torch.LongTensor([0]), end_channelids[:-1]])
        dst_ls = torch.repeat_interleave(ls_metadata, n_irreps_per_lp * metadata1d)
        dst_ms = torch.repeat_interleave(
            torch.cat([torch.arange(-l, l + 1) for l in ls_metadata]),
            torch.repeat_interleave(metadata1d, n_irreps_per_lp),
        )
        ns = torch.arange(metadata1d.sum())
        dst_ns = torch.cat(
            [
                ns[start_channelids[l] : end_channelids[l]].repeat(n_irreps)
                for l, n_irreps in enumerate(n_irreps_per_lp)
            ]
        )
        ps = torch.LongTensor([1, -1]).repeat(num_distinct_ls)
        dst_ps = torch.repeat_interleave(ps, n_irreps_per_lp * metadata1d)
        rep_layout = torch.stack([dst_ls, dst_ms, dst_ns, dst_ps], dim=0).long()
        assert rep_layout.shape[1] == torch.sum(n_irreps_per_lp * metadata1d)

        return rep_layout

    @classmethod
    def from_so3(cls, so3_ten: SphericalTensor, parity=1):
        metadata = so3_ten.metadata
        if parity == 1:
            o3_metadata = torch.stack(
                [metadata, torch.zeros_like(metadata)], dim=2
            ).view(metadata.shape[0], -1)
            o3_layout = tuple(
                torch.cat(
                    [
                        layout,
                        torch.ones(
                            1, layout.shape[1], dtype=torch.long, device=layout.device
                        ),
                    ],
                    dim=0,
                )
                for layout in so3_ten.rep_layout
            )
        elif parity == -1:
            o3_metadata = torch.stack(
                [torch.zeros_like(metadata), metadata], dim=2
            ).view(metadata.shape[0], -1)
            o3_layout = tuple(
                torch.cat(
                    [
                        layout,
                        torch.ones(
                            1, layout.shape[1], dtype=torch.long, device=layout.device
                        ).neg(),
                    ],
                    dim=0,
                )
                for layout in so3_ten.rep_layout
            )
        else:
            raise NotImplementedError

        return O3Tensor(
            so3_ten.ten,
            rep_dims=so3_ten.rep_dims,
            metadata=o3_metadata,
            rep_layout=o3_layout,
            num_channels=so3_ten.num_channels,
        )



import math
import random

import torch

from torch_gauge.o3 import O3Tensor
from torch_gauge.o3.clebsch_gordan import CGPCoupler
from torch_gauge.o3.wigner import wigner_D_rsh

torch.manual_seed(42)


def test_cgp_selection_rule():
    # max_l = 1, same layout
    metadata = torch.LongTensor([[24, 0, 12, 0]])
    dten1 = torch.rand(4, 6, 2, 60, 7)
    dten2 = torch.rand(4, 6, 2, 60, 7)
    spten1 = O3Tensor(dten1, (3,), metadata)
    spten2 = O3Tensor(dten2, (3,), metadata)
    coupler2 = CGPCoupler(metadata[0], metadata[0], trunc_in=False, dtype=torch.float)
    cat_out = coupler2(spten1, spten2)
    assert cat_out.ten.shape == (4, 6, 2, 108, 7)
    assert torch.all(cat_out.metadata.eq(torch.LongTensor([[36, 0, 24, 0]])))
    coupler3 = CGPCoupler(metadata[0], metadata[0], trunc_in=True, dtype=torch.float)
    trunc_out = coupler3(spten1, spten2)
    assert trunc_out.ten.shape == (4, 6, 2, 96, 7)
    assert torch.all(trunc_out.metadata.eq(torch.LongTensor([[24, 0, 24, 0]])))

    # max_l = 1, hetero layout
    metadata1 = torch.LongTensor([[24, 0, 12, 0]])
    metadata2 = torch.LongTensor([[17, 3, 5, 15]])
    dten1 = torch.rand(4, 6, 2, 60, 7)
    dten2 = torch.rand(4, 6, 2, 80, 7)
    spten1 = O3Tensor(dten1, (3,), metadata1)
    spten2 = O3Tensor(dten2, (3,), metadata2)
    coupler2 = CGPCoupler(metadata1[0], metadata2[0], trunc_in=False, dtype=torch.float)
    cat_out = coupler2(spten1, spten2)
    assert torch.all(cat_out.metadata.eq(torch.LongTensor([[22, 6, 29, 23]])))
    assert cat_out.ten.shape == (4, 6, 2, 184, 7)
    coupler3 = CGPCoupler(metadata1[0], metadata2[0], trunc_in=True, dtype=torch.float)
    trunc_out = coupler3(spten1, spten2)
    assert torch.all(trunc_out.metadata.eq(torch.LongTensor([[17, 3, 17, 18]])))
    assert trunc_out.ten.shape == (4, 6, 2, 125, 7)

    # Large l coupling
    metadata1 = torch.LongTensor(
        [[128, 64, 64, 32, 32, 16, 16, 8, 8, 4, 4, 2, 2, 1, 1, 0, 1, 0]]
    )
    metadata2 = torch.LongTensor(
        [[160, 80, 80, 40, 40, 20, 20, 10, 10, 5, 5, 1, 1, 0, 0, 0, 0, 0]]
    )
    dten1 = torch.rand(8, 1133)
    dten2 = torch.rand(8, 1324)
    spten1 = O3Tensor(dten1, (1,), metadata1)
    spten2 = O3Tensor(dten2, (1,), metadata2)
    coupler2 = CGPCoupler(metadata1[0], metadata2[0], trunc_in=False, dtype=torch.float)
    coupler3 = CGPCoupler(metadata1[0], metadata2[0], trunc_in=True, dtype=torch.float)

    coupling_out = coupler2(spten1, spten2)
    assert torch.all(
        coupling_out.metadata.eq(
            torch.LongTensor(
                [
                    [
                        378,
                        283,
                        561,
                        503,
                        558,
                        455,
                        438,
                        339,
                        306,
                        219,
                        188,
                        106,
                        89,
                        59,
                        48,
                        0,
                        37,
                        0,
                    ]
                ]
            )
        )
    )
    assert not torch.any(coupling_out.ten == 0)
    coupling_out = coupler3(spten1, spten2)
    assert torch.all(
        coupling_out.metadata.eq(
            torch.LongTensor(
                [
                    [
                        290,
                        196,
                        414,
                        252,
                        356,
                        200,
                        253,
                        130,
                        160,
                        75,
                        90,
                        33,
                        44,
                        13,
                        22,
                        0,
                        11,
                        0,
                    ]
                ]
            )
        )
    )
    assert not torch.any(coupling_out.ten == 0)


metadata1 = torch.LongTensor(
    [[128, 64, 64, 32, 32, 16, 16, 8, 8, 4, 4, 2, 2, 1, 1, 0, 1, 0]]
)
metadata2 = torch.LongTensor(
    [[160, 80, 80, 40, 40, 20, 20, 10, 10, 5, 5, 1, 1, 0, 0, 0, 0, 0]]
)
dten1 = torch.rand(48, 1133)
dten2 = torch.rand(48, 1324)
spten1 = O3Tensor(dten1, (1,), metadata1)
spten2 = O3Tensor(dten2, (1,), metadata2)
coupler1 = CGPCoupler(metadata1[0], metadata2[0], trunc_in=False, dtype=torch.float)
coupler2 = CGPCoupler(metadata1[0], metadata2[0], trunc_in=True, dtype=torch.float)


def test_benchmark_cgpcoupler_setup1(benchmark):
    benchmark(coupler1, spten1, spten2)


def test_benchmark_cgpcoupler_setup2(benchmark):
    benchmark(coupler2, spten1, spten2)


def real_wigner_rot(o3ten, alpha, beta, gamma):
    # Naive implementation, do not use in models
    rot_ten = torch.zeros_like(o3ten.ten)
    for l in range(o3ten.metadata.shape[1] // 2):
        for p in (1, -1):
            print(f"Rotating rep {l}, parity {p}...")
            real_wigner_l = wigner_D_rsh(l, alpha, beta, gamma).type(o3ten.ten.dtype)
            l_mask = o3ten.rep_layout[0][0] == l
            p_mask = o3ten.rep_layout[0][3] == p
            torot = o3ten.ten[:, l_mask * p_mask]
            if torot.numel() == 0:
                continue
            torot = torot.view(o3ten.ten.shape[0], 2 * l + 1, -1)
            rot_ten[:, l_mask * p_mask] = (
                torch.einsum("ijk,ja->iak", torot, real_wigner_l)
                .contiguous()
                .view(torot.shape[0], -1)
            )
    return o3ten.self_like(rot_ten)


def condon_shortley_inv(o3ten):
    # Total parity = (-1)**l * p
    cs_phase = torch.pow(-1, o3ten.rep_layout[0][0]) * o3ten.rep_layout[0][3]
    return o3ten.self_like(o3ten.ten.mul(cs_phase))


# Equivariance tests w.r.t improper rotations
alpha = random.random() * 2 * math.pi
beta = random.random() * 2 * math.pi
gamma = random.random() * 2 * math.pi


def test_cgp_coupling_equivariance1():
    out1: O3Tensor = coupler1(spten1, spten2)
    out1_rot = condon_shortley_inv(real_wigner_rot(out1, alpha, beta, gamma))

    spten1_rot = condon_shortley_inv(real_wigner_rot(spten1, alpha, beta, gamma))
    spten2_rot = condon_shortley_inv(real_wigner_rot(spten2, alpha, beta, gamma))
    rot_out1 = coupler1(spten1_rot, spten2_rot)
    assert not torch.allclose(out1_rot.ten, out1.ten, atol=1e-3), print(
        (out1_rot.ten - out1.ten).flatten().abs()
    )
    print(out1.ten)
    print(out1_rot.ten)
    print(rot_out1.ten)
    assert torch.allclose(rot_out1.ten, out1_rot.ten, atol=1e-5, rtol=1e-5), print(
        (out1_rot.ten - rot_out1.ten).flatten().abs(), out1_rot.metadata
    )


def test_cgp_coupling_equivariance2():
    out2: O3Tensor = coupler2(spten1, spten2)
    out2_rot = condon_shortley_inv(real_wigner_rot(out2, alpha, beta, gamma))

    spten1_rot = condon_shortley_inv(real_wigner_rot(spten1, alpha, beta, gamma))
    spten2_rot = condon_shortley_inv(real_wigner_rot(spten2, alpha, beta, gamma))
    rot_out2 = coupler2(spten1_rot, spten2_rot)
    assert not torch.allclose(out2_rot.ten, out2.ten, atol=1e-3), print(
        (out2_rot.ten - out2.ten).flatten().abs()
    )
    assert torch.allclose(rot_out2.ten, out2_rot.ten, atol=1e-5, rtol=1e-5), print(
        (out2_rot.ten - rot_out2.ten).flatten().abs()
    )



import torch

from torch_gauge import VerletList
from torch_gauge.o3 import O3Tensor


# The i-j inversion should return a transposed matrix when the graph is dense
def test_verletlist_inversion_dense():
    data = torch.rand(17, 17, 8, 6)
    mask = torch.ones(17, 17, dtype=torch.bool)
    vl_test = VerletList().from_mask(mask, 17, 17, {}, {"test_edata": data})
    inversed = vl_test.to_src_first_view(vl_test.edata["test_edata"])
    assert torch.all(data.permute(1, 0, 2, 3).eq(inversed))


def test_verletlist_inversion_sparse():
    data = torch.rand(17, 17, 8, 6)
    mask_pre = torch.rand(17, 17)
    mask = (mask_pre + mask_pre.t()) > 0.5
    vl_test = VerletList().from_mask(mask, 32, 17, {}, {"test_edata": data})
    assert torch.all(
        data[mask, ...].eq(vl_test.edata["test_edata"][vl_test.edge_mask, ...])
    )

    inversed = vl_test.to_src_first_view(vl_test.edata["test_edata"])
    sparse1 = torch.zeros_like(data)
    sparse1[mask, ...] = data[mask, ...]
    sparse2 = torch.zeros_like(data)
    sparse2[mask, ...] = inversed[vl_test.edge_mask, ...]
    assert torch.all(sparse1.permute(1, 0, 2, 3).eq(sparse2))


# When the graph is dense, query_src inserts a dimension
def test_verletlist_querysrc_dense():
    data = torch.rand(17, 8, 6)
    mask = torch.ones(17, 17, dtype=torch.bool)
    vl_test = VerletList().from_mask(mask, 17, 17, {"test_ndata": data}, {})
    src_data = vl_test.query_src(vl_test.ndata["test_ndata"])
    assert torch.all(data.unsqueeze(0).expand(17, 17, 8, 6).eq(src_data))


def test_verletlist_o3():
    vls = []
    for _ in range(4):
        dten1 = torch.rand(17, 101, 7)
        metadata = torch.LongTensor([[3, 4, 5, 4, 1, 0, 2, 3, 2, 1]])
        test1 = O3Tensor(dten1, (1,), metadata)
        dten2 = torch.rand(17, 17, 101, 101, 7)
        test2 = O3Tensor(dten2, (2, 3), metadata.repeat(2, 1))
        mask_pre = torch.rand(17, 17)
        mask = (mask_pre + mask_pre.t()) > 0.5
        vl_test = VerletList().from_mask(
            mask, 17, 17, {"test_ndata": test1}, {"test_edata": test2}
        )
        vls.append(vl_test)
    batched_vls = VerletList.batch(vls)
    assert batched_vls.ndata["test_ndata"].rep_dims == (1,)
    assert batched_vls.ndata["test_ndata"].ten.shape == (17 * 4, 101, 7)
    assert batched_vls.edata["test_edata"].rep_dims == (
        2,
        3,
    )
    assert batched_vls.edata["test_edata"].ten.shape == (17 * 4, 17, 101, 101, 7)
    expanded_n = batched_vls.query_src(batched_vls.ndata["test_ndata"])
    expanded_e = batched_vls.query_src(batched_vls.edata["test_edata"])
    assert expanded_n.rep_dims == (2,)
    assert expanded_n.ten.shape == (17 * 4, 17, 101, 7)
    assert expanded_e.rep_dims == (3, 4)
    assert expanded_e.ten.shape == (17 * 4, 17, 17, 101, 101, 7)
    inverted_e = batched_vls.to_src_first_view(batched_vls.edata["test_edata"])
    assert inverted_e.rep_dims == (2, 3)
    assert inverted_e.ten.shape == (17 * 4, 17, 101, 101, 7)


if __name__ == "__main__":
    test_verletlist_inversion_dense()
    test_verletlist_querysrc_dense()



"""
Testing Spherical neural operations
"""

import math
import random

import torch

from torch_gauge.nn import IELin, KernelBroadcast, RepNorm1d
from torch_gauge.o3 import SphericalTensor
from torch_gauge.o3.rsh import RSHxyz
from torch_gauge.o3.wigner import wigner_D_rsh


def real_wigner_rot(spten, alpha, beta, gamma):
    # Naive implementation, do not use in models
    rot_ten = torch.zeros_like(spten.ten)
    for l in range(spten.metadata.shape[1]):
        real_wigner_l = wigner_D_rsh(l, alpha, beta, gamma).type(spten.ten.dtype)
        l_mask = spten.rep_layout[0][0] == l
        torot = spten.ten[:, l_mask]
        if torot.numel() == 0:
            continue
        torot = torot.view(spten.ten.shape[0], 2 * l + 1, -1)
        rot_ten[:, l_mask] = (
            torch.einsum("ijk,ja->iak", torot, real_wigner_l)
            .contiguous()
            .view(torot.shape[0], -1)
        )
    return spten.self_like(rot_ten)


def test_IELin_forward_1d():
    metadata = torch.LongTensor([[8, 4, 2, 4, 1]])
    test_sp_ten = SphericalTensor(torch.rand(4, 6, 12, 67), (3,), metadata)
    ielin_dense = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 3, 4, 5, 6]),
        group="so3",
    )
    dense_out = ielin_dense(test_sp_ten)
    assert dense_out.ten.shape == (4, 6, 12, 120)
    assert dense_out.num_channels == (20,)
    assert torch.all(dense_out.metadata.eq(torch.LongTensor([[2, 3, 4, 5, 6]])))

    ielin_dropped = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 0, 1, 3, 0]),
        group="so3",
    )
    dropped_out = ielin_dropped(test_sp_ten)
    assert dropped_out.ten.shape == (4, 6, 12, 28)
    assert dropped_out.num_channels == (6,)
    assert torch.all(dropped_out.metadata.eq(torch.LongTensor([[2, 0, 1, 3, 0]])))


def test_IELin_forward_equivariance():
    alpha = random.random() * 2 * math.pi
    beta = random.random() * 2 * math.pi
    gamma = random.random() * 2 * math.pi
    metadata = torch.LongTensor([[8, 4, 2, 4, 1]])
    test_sp_ten = SphericalTensor(torch.rand(48, 67), (1,), metadata)
    spten_rot = real_wigner_rot(test_sp_ten, alpha, beta, gamma)
    ielin_dense = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 3, 4, 5, 6]),
        group="so3",
    )
    dense_out = ielin_dense(test_sp_ten)
    rot_dense_out = real_wigner_rot(dense_out, alpha, beta, gamma)
    dense_out_rot = ielin_dense(spten_rot)
    assert torch.allclose(
        rot_dense_out.ten, dense_out_rot.ten, atol=1e-6, rtol=1e-6
    ), print((dense_out_rot.ten - rot_dense_out.ten).flatten().abs())

    ielin_dropped = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 0, 1, 3, 0]),
        group="so3",
    )
    dropped_out = ielin_dropped(test_sp_ten)
    dropped_out_rot = ielin_dropped(spten_rot)
    rot_dropped_out = real_wigner_rot(dropped_out, alpha, beta, gamma)
    assert torch.allclose(
        rot_dropped_out.ten, dropped_out_rot.ten, atol=1e-6, rtol=1e-6
    ), print((rot_dropped_out.ten - dropped_out_rot.ten).flatten().abs())


def test_IELin_backward_1d():
    metadata = torch.LongTensor([[8, 4, 2, 4, 1]])
    test_sp_ten = SphericalTensor(torch.rand(4, 6, 12, 67), (3,), metadata)
    test_sp_ten.ten.requires_grad = True
    ielin_dense = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 3, 4, 5, 6]),
        group="so3",
    )
    dense_out = ielin_dense(test_sp_ten)
    pseudo_loss = dense_out.ten.pow(2).sum()
    grad_ten = torch.autograd.grad(outputs=pseudo_loss, inputs=test_sp_ten.ten)
    assert torch.all(grad_ten[0].abs().bool())

    ielin_dropped = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 0, 1, 3, 0]),
        group="so3",
    )
    dropped_out = ielin_dropped(test_sp_ten)
    pseudo_loss = dropped_out.ten.pow(2).sum()
    grad_ten = torch.autograd.grad(
        outputs=pseudo_loss, inputs=test_sp_ten.ten, retain_graph=True
    )
    assert torch.any(grad_ten[0].abs().bool())
    assert not torch.all(grad_ten[0].abs().bool())
    pseudo_loss.backward()
    assert torch.all(ielin_dropped.linears.grad.abs().sum((1, 2)).bool())

    # Test Norm contraction backwards
    test_sp_ten.ten.grad.data.zero_()
    ielin_dropped.linears.grad.data.zero_()
    dropped_out = ielin_dropped(test_sp_ten)
    pseudo_loss = dropped_out.invariant().sum()
    grad_ten = torch.autograd.grad(
        outputs=pseudo_loss, inputs=test_sp_ten.ten, retain_graph=True
    )
    assert torch.any(grad_ten[0].abs().bool())
    assert not torch.all(grad_ten[0].abs().bool())
    pseudo_loss.backward()
    assert torch.all(ielin_dropped.linears.grad.abs().sum((1, 2)).bool())


def test_IELin_forward_2d():
    metadata = torch.LongTensor([[8, 4, 2, 4, 1], [8, 4, 2, 4, 1]])
    test_sp_ten = SphericalTensor(
        torch.rand(4, 6, 12, 67, 67),
        (
            3,
            4,
        ),
        metadata,
    )
    ielin_dense = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 3, 4, 5, 6]),
        group="so3",
    )
    out1 = ielin_dense(test_sp_ten)
    assert out1.ten.shape == (4, 6, 12, 67, 120)
    assert out1.num_channels == (
        19,
        20,
    )
    assert torch.all(
        out1.metadata.eq(torch.LongTensor([[8, 4, 2, 4, 1], [2, 3, 4, 5, 6]]))
    )

    out2 = ielin_dense(out1.transpose_repdims(inplace=True))
    assert out2.ten.shape == (4, 6, 12, 120, 120)
    assert out2.num_channels == (
        20,
        20,
    )
    assert torch.all(
        out2.metadata.eq(torch.LongTensor([[2, 3, 4, 5, 6], [2, 3, 4, 5, 6]]))
    )


def test_IELin_backward_2d():
    metadata = torch.LongTensor([[8, 4, 2, 4, 1], [8, 4, 2, 4, 1]])
    test_sp_ten = SphericalTensor(
        torch.rand(4, 6, 12, 67, 67),
        (
            3,
            4,
        ),
        metadata,
    )
    test_sp_ten.ten.requires_grad = True
    ielin_dense = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 3, 4, 5, 6]),
        group="so3",
    )
    out1 = ielin_dense(test_sp_ten)
    out2 = ielin_dense(out1.transpose_repdims(inplace=True))

    pseudo_loss = out2.ten.pow(2).sum()
    grad_ten = torch.autograd.grad(
        outputs=pseudo_loss, inputs=test_sp_ten.ten, retain_graph=True
    )
    assert grad_ten[0].shape == (4, 6, 12, 67, 67)
    assert torch.any(grad_ten[0].abs().bool())
    assert torch.all(grad_ten[0].abs().bool())
    pseudo_loss.backward()
    assert torch.all(ielin_dense.linears.grad.abs().sum((1, 2)).bool())

    # Test Norm contraction backwards
    test_sp_ten.ten.grad.data.zero_()
    ielin_dense.linears.grad.data.zero_()
    out1 = ielin_dense(test_sp_ten)
    out2 = ielin_dense(out1.transpose_repdims(inplace=True))
    pseudo_loss = out2.invariant().sum() - 0.3
    grad_ten = torch.autograd.grad(
        outputs=pseudo_loss, inputs=test_sp_ten.ten, retain_graph=True
    )
    assert grad_ten[0].shape == (4, 6, 12, 67, 67)
    assert torch.any(grad_ten[0].abs().bool())
    assert torch.all(grad_ten[0].abs().bool())
    pseudo_loss.backward()
    assert torch.all(ielin_dense.linears.grad.abs().sum((1, 2)).bool())


def test_RepNorm1d_forward():
    metadata = torch.LongTensor([[8, 4, 2, 4, 1]])
    test_sp_batched = SphericalTensor(torch.rand(1024, 67), (1,), metadata)
    activation = torch.nn.Tanh()
    norm_linear = torch.nn.Linear(19, 20)
    gauge_linear = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 3, 4, 5, 6]),
        group="so3",
    )
    repnorm = RepNorm1d(19)

    n, g = repnorm(test_sp_batched)
    out = gauge_linear(g).scalar_mul(activation(norm_linear(n)), inplace=True)
    assert out.ten.shape == (1024, 120)
    assert torch.all(out.metadata.eq(torch.LongTensor([[2, 3, 4, 5, 6]])))


def test_RepNorm1d_backward():
    metadata = torch.LongTensor([[8, 4, 2, 4, 1]])
    test_sp_batched = SphericalTensor(torch.rand(1024, 67), (1,), metadata)
    test_sp_batched.ten.requires_grad = True
    activation = torch.nn.Tanh()
    norm_linear = torch.nn.Linear(19, 20)
    gauge_linear = IELin(
        torch.LongTensor([8, 4, 2, 4, 1]),
        torch.LongTensor([2, 3, 4, 5, 6]),
        group="so3",
    )
    repnorm = RepNorm1d(19)

    n, g = repnorm(test_sp_batched)
    out = gauge_linear(g).scalar_mul(activation(norm_linear(n)), inplace=True)

    pseudo_loss = out.invariant().pow(2).sum() - 0.3
    pseudo_loss.backward()
    assert torch.all(test_sp_batched.ten.grad.abs().bool())
    assert torch.all(norm_linear.weight.grad.abs().bool())
    assert torch.all(norm_linear.bias.grad.abs().bool())
    assert torch.all(gauge_linear.linears.grad.abs().sum((1, 2)).bool())


def test_kernel_broadcast():
    rshmodule = RSHxyz(max_l=6)
    xyz = torch.normal(0, 1, size=(64, 3))
    rshs = rshmodule(xyz)
    feat = torch.rand(64, 128)
    kbd = KernelBroadcast(torch.LongTensor([64, 32, 16, 8, 4, 2, 2]))
    out = kbd(rshs, feat)
    assert torch.all(out != 0)



"""
Geometric operations on point-clouds and manifolds.
"""
import math

import torch


class UnivecAngle(torch.autograd.Function):
    """
    Backward-robust angle calculation between two unit vectors.

    Note:
        For possible evaluation schemes, see:
            https://scicomp.stackexchange.com/questions/27689/numerically-stable-way-of-computing-angles-between-vectors

        Also see:
            https://www.cs.utexas.edu/users/evouga/uploads/4/5/6/8/45689883/turning.pdf
            which is algebraically simpler but involves more forward arithmetic operations.

        On the benchmark CPU, tensor.norm() is much slower than dot and 2x slower than
        cross product. We take the second approach here.
    """

    @staticmethod
    def forward(ctx, vec1, vec2):
        """"""
        cross_vec = torch.cross(vec1, vec2, dim=-1)
        cross_norm = cross_vec.pow(2).sum(dim=-1).sqrt()
        vec_dot = vec1.mul(vec2).sum(-1)
        angle = torch.atan2(cross_norm, 1 + vec_dot).mul_(2)
        ctx.save_for_backward(vec1, vec2, cross_vec, cross_norm)
        return angle

    @staticmethod
    def backward(ctx, grad_output):
        """"""
        vec1, vec2, cross_vec, cross_norm = ctx.saved_tensors
        cross_norm = cross_norm.unsqueeze(-1).expand_as(cross_vec)
        z = cross_vec.div(cross_norm)
        # Need to consider using explicit Chebyshev-poly if this breaks in near-edge cases
        z.masked_fill_(cross_norm == 0, 0)
        grad_vec1, grad_vec2 = torch.cross(vec1, z), torch.cross(vec2, z).neg_()
        return grad_vec1.mul(grad_output.unsqueeze(-1)), grad_vec2.mul(
            grad_output.unsqueeze(-1)
        )


def univec_angle(vec1: torch.Tensor, vec2: torch.Tensor):
    return UnivecAngle.apply(vec1, vec2)


def univec_angle_unsafe(vec1: torch.Tensor, vec2: torch.Tensor):
    return torch.acos(vec1.mul(vec2).sum(-1))


def univec_cos(vec1: torch.Tensor, vec2: torch.Tensor):
    """
    Returns :math:`\cos(\theta)` for backward-robust angle calculations.
    """
    return vec1.mul(vec2).sum(-1)


def rotation_matrix_xyz(phi, axis, dtype=torch.double):
    """
    Rotation matrix for the intrinsic rotation about the specified axis,
    with the column-xyz (left-multiplication) convention.
    """
    if axis == "z":
        return torch.tensor(
            [
                [math.cos(phi), -math.sin(phi), 0],
                [math.sin(phi), math.cos(phi), 0],
                [0, 0, 1],
            ],
            dtype=dtype,
        )
    elif axis == "y":
        return torch.tensor(
            [
                [math.cos(phi), 0, math.sin(phi)],
                [0, 1, 0],
                [-math.sin(phi), 0, math.cos(phi)],
            ],
            dtype=dtype,
        )
    elif axis == "x":
        return torch.tensor(
            [
                [1, 0, 0],
                [0, math.cos(phi), -math.sin(phi)],
                [0, math.sin(phi), math.cos(phi)],
            ],
            dtype=dtype,
        )
    else:
        raise ValueError(f"Invalid axis: {axis}")


class Chebyshev(torch.nn.Module):
    """The Module to generate Chebshev polynomials up to the given order."""

    def __init__(self, order):
        super(Chebyshev, self).__init__()
        self.order = order
        self.orders = torch.arange(order + 1, dtype=torch.long)
        self.orders = torch.nn.Parameter(self.orders, requires_grad=False)
        self._init_coefficients()

    def _init_coefficients(self):
        """
        Generating combination coefficients using recursion,
        starting from order 1.
        """
        coeff = torch.zeros(self.order + 1, self.order + 1)
        coeff[0, 0] = 1
        coeff[1, 1] = 1
        for idx in range(2, self.order + 1):
            coeff[1:, idx] = 2 * coeff[:-1, idx - 1]
            coeff[:, idx] -= coeff[:, idx - 2]
        self.cheb_coeff = torch.nn.Parameter(coeff[:, 1:], requires_grad=False)

    def forward(self, x):
        """"""
        size = (*x.shape, self.order + 1)
        x = x.unsqueeze(-1).expand(size).pow(self.orders)
        out = torch.matmul(x, self.cheb_coeff)
        return out


def poly_env(d, p=6):
    """
    A polynomial wrapper defined in:

        J. Klicpera, J. Groß, and S. G¨unnemann, arXiv preprint arXiv:2003.03123 (2020).
    """
    return (
        1
        - (p + 1) * (p + 2) // 2 * d ** p
        + p * (p + 2) * d ** (p + 1)
        - p * (p + 1) // 2 * d ** (p + 2)
    )



"""
Test training a minimal SO(3) model on synthetic data
"""

import pytest
import torch

from torch_gauge.nn import IELin, RepNorm1d, Swish
from torch_gauge.o3 import O3Tensor, SphericalTensor
from torch_gauge.o3.clebsch_gordan import CGPCoupler


class mini2d_so3(torch.nn.Module):
    def __init__(self, metadata, n_channels):
        super().__init__()
        self.onebody_ielin = IELin(metadata, metadata, group="so3")
        self.onebody_mlp = torch.nn.Sequential(
            torch.nn.Linear(n_channels, n_channels),
            Swish(),
            torch.nn.Linear(n_channels, n_channels),
        )
        self.onebody_norm = RepNorm1d(num_channels=n_channels)
        self.twobody_linear = torch.nn.Linear(4, 4, bias=False)
        self.twobody_gate = torch.nn.Sequential(
            torch.nn.Linear(4, 4),
            Swish(),
        )

    def forward(self, x_1d: SphericalTensor, x_2d: SphericalTensor):
        n, g = self.onebody_norm(x_1d)
        n = self.onebody_mlp(n)
        g = self.onebody_ielin(g)
        x_1d_new = g.scalar_mul(n, inplace=True)
        fold_1d = x_1d_new.fold(stride=4)
        a = (
            x_2d.ten.mul(fold_1d.ten.unsqueeze(1))
            .sum(dim=2)
            .mul(fold_1d.ten)
            .sum(dim=1)
        )
        a = self.twobody_gate(a)
        new_2d_ten = (self.twobody_linear(x_2d.ten)).mul(a.unsqueeze(1).unsqueeze(2))
        x_2d_new = x_2d.self_like(new_2d_ten)
        return x_1d_new, x_2d_new


def test_train_cpu_mini2d_so3():
    torch.manual_seed(42)
    metadata1d = torch.LongTensor([[12, 4, 8, 4, 4]])
    metadata2d = torch.LongTensor([[3, 1, 2, 1, 1], [3, 1, 2, 1, 1]])
    x_1d = SphericalTensor(torch.rand(128, 128), (1,), metadata1d)
    x_2d = SphericalTensor(torch.rand(128, 32, 32, 4), (1, 2), metadata2d)
    # Create synthetic labels
    rs_1d = x_1d.ten.view(128, 32, 4)
    labels = torch.sin(
        (x_2d.ten * rs_1d.unsqueeze(1) * rs_1d.unsqueeze(2))
        .pow(2)
        .mean(dim=(1, 2))
        .sqrt()
    ).sum(1)

    mods = torch.nn.ModuleList(
        [
            mini2d_so3(torch.LongTensor([12, 4, 8, 4, 4]), 32),
            mini2d_so3(torch.LongTensor([12, 4, 8, 4, 4]), 32),
            mini2d_so3(torch.LongTensor([12, 4, 8, 4, 4]), 32),
        ]
    )

    loss_fn = torch.nn.MSELoss()
    mae_fn = torch.nn.L1Loss()
    optimizer = torch.optim.Adam(mods.parameters(), lr=1e-3)

    loss, epoch_mae = 0.0, 0.0
    for epoch in range(50):
        epoch_mae = 0.0
        for iter in range(4):
            batch_idx = torch.randint(128, (32,))
            batch_label = labels[batch_idx]
            x_1d_new = x_1d.self_like(x_1d.ten[batch_idx])
            x_2d_new = x_2d.self_like(x_2d.ten[batch_idx])
            for mod in mods:
                x_1d_new, x_2d_new = mod(x_1d_new, x_2d_new)
            out = x_2d_new.invariant().sum(dim=(1, 2, 3))
            optimizer.zero_grad()
            loss = loss_fn(out, batch_label)
            mae = mae_fn(out, batch_label)
            loss.backward()
            optimizer.step()
            epoch_mae += mae.item()
        print(f"Epoch: {epoch+1}, Loss: {loss}, MAE: {epoch_mae/4}")

    assert loss < 0.01
    assert epoch_mae / 4 < 0.1


class mini2d_o3(torch.nn.Module):
    def __init__(self, metadata, n_channels):
        super().__init__()
        self.onebody_ielin1 = IELin(metadata, metadata, group="o3")
        self.onebody_mlp = torch.nn.Sequential(
            torch.nn.Linear(n_channels, n_channels),
            Swish(),
            torch.nn.Linear(n_channels, n_channels),
        )
        self.onebody_norm = RepNorm1d(
            num_channels=n_channels, n_invariant_channels=metadata[0]
        )
        self.onebody_coupling = CGPCoupler(metadata, metadata, dtype=torch.float)
        self.onebody_ielin2 = IELin(
            self.onebody_coupling.metadata_out, metadata, group="o3"
        )
        self.twobody_linear = torch.nn.Linear(4, 4, bias=False)
        self.twobody_gate = torch.nn.Sequential(
            torch.nn.Linear(4, 4),
            Swish(),
        )

    def forward(self, x_1d: O3Tensor, x_2d: O3Tensor):
        n, g = self.onebody_norm(x_1d)
        n = self.onebody_mlp(n)
        g = self.onebody_ielin1(g)
        g_cgp = self.onebody_coupling(g, g)
        g = self.onebody_ielin2(g_cgp) + g
        x_1d_new = g.scalar_mul(n, inplace=True)
        fold_1d = x_1d_new.fold(stride=4)
        a = (
            x_2d.ten.mul(fold_1d.ten.unsqueeze(1))
            .sum(dim=2)
            .mul(fold_1d.ten)
            .sum(dim=1)
        )
        a = self.twobody_gate(a)
        new_2d_ten = (self.twobody_linear(x_2d.ten)).mul(a.unsqueeze(1).unsqueeze(2))
        x_2d_new = x_2d.self_like(new_2d_ten)
        return x_1d_new, x_2d_new


def test_train_cpu_mini2d_o3():
    torch.manual_seed(42)
    metadata1d = torch.LongTensor([[12, 4, 8, 4, 4, 4]])
    metadata2d = torch.LongTensor([[3, 1, 2, 1, 1, 1], [3, 1, 2, 1, 1, 1]])
    x_1d = O3Tensor(torch.rand(128, 92), (1,), metadata1d)
    x_2d = O3Tensor(torch.rand(128, 23, 23, 4), (1, 2), metadata2d)
    # Create synthetic labels
    rs_1d = x_1d.ten.view(128, 23, 4)
    labels = torch.sin(
        (x_2d.ten * rs_1d.unsqueeze(1) * rs_1d.unsqueeze(2))
        .pow(2)
        .mean(dim=(1, 2))
        .sqrt()
    ).sum(1)

    mods = torch.nn.ModuleList(
        [
            mini2d_o3(torch.LongTensor([12, 4, 8, 4, 4, 4]), 36),
            mini2d_o3(torch.LongTensor([12, 4, 8, 4, 4, 4]), 36),
            mini2d_o3(torch.LongTensor([12, 4, 8, 4, 4, 4]), 36),
        ]
    )

    loss_fn = torch.nn.MSELoss()
    mae_fn = torch.nn.L1Loss()
    optimizer = torch.optim.Adam(mods.parameters(), lr=1e-3)

    loss, epoch_mae = 0.0, 0.0
    for epoch in range(50):
        epoch_mae = 0.0
        for iter in range(4):
            batch_idx = torch.randint(128, (32,))
            batch_label = labels[batch_idx]
            x_1d_new = x_1d.self_like(x_1d.ten[batch_idx])
            x_2d_new = x_2d.self_like(x_2d.ten[batch_idx])
            for mod in mods:
                x_1d_new, x_2d_new = mod(x_1d_new, x_2d_new)
            out = x_2d_new.invariant().sum(dim=(1, 2, 3))
            optimizer.zero_grad()
            loss = loss_fn(out, batch_label)
            mae = mae_fn(out, batch_label)
            loss.backward()
            optimizer.step()
            epoch_mae += mae.item()
        print(f"Epoch: {epoch+1}, Loss: {loss}, MAE: {epoch_mae/4}")

    assert loss < 0.02
    assert epoch_mae / 4 < 0.1


def test_train_cuda_mini2d_o3():
    if not torch.cuda.is_available():
        pytest.skip()
    device = torch.device("cuda")
    torch.manual_seed(42)
    metadata1d = torch.LongTensor([[12, 4, 8, 4, 4, 4]])
    metadata2d = torch.LongTensor([[3, 1, 2, 1, 1, 1], [3, 1, 2, 1, 1, 1]])
    x_1d = O3Tensor(torch.rand(128, 92), (1,), metadata1d).to(device)
    x_2d = O3Tensor(torch.rand(128, 23, 23, 4), (1, 2), metadata2d).to(device)
    # Create synthetic labels
    rs_1d = x_1d.ten.view(128, 23, 4)
    labels = torch.sin(
        (x_2d.ten * rs_1d.unsqueeze(1) * rs_1d.unsqueeze(2))
        .pow(2)
        .mean(dim=(1, 2))
        .sqrt()
    ).sum(1)

    mods = torch.nn.ModuleList(
        [
            mini2d_o3(torch.LongTensor([12, 4, 8, 4, 4, 4]), 36),
            mini2d_o3(torch.LongTensor([12, 4, 8, 4, 4, 4]), 36),
            mini2d_o3(torch.LongTensor([12, 4, 8, 4, 4, 4]), 36),
        ]
    ).to(device)

    loss_fn = torch.nn.MSELoss()
    mae_fn = torch.nn.L1Loss()
    optimizer = torch.optim.Adam(mods.parameters(), lr=1e-3)

    loss, epoch_mae = 0.0, 0.0
    for epoch in range(50):
        epoch_mae = 0.0
        for iter in range(4):
            batch_idx = torch.randint(128, (32,))
            batch_label = labels[batch_idx]
            x_1d_new = x_1d.self_like(x_1d.ten[batch_idx])
            x_2d_new = x_2d.self_like(x_2d.ten[batch_idx])
            for mod in mods:
                x_1d_new, x_2d_new = mod(x_1d_new, x_2d_new)
            out = x_2d_new.invariant().sum(dim=(1, 2, 3))
            optimizer.zero_grad()
            loss = loss_fn(out, batch_label)
            mae = mae_fn(out, batch_label)
            loss.backward()
            optimizer.step()
            epoch_mae += mae.cpu().item()
        print(f"Epoch: {epoch+1}, Loss: {loss.cpu().item()}, MAE: {epoch_mae/4}")

    assert loss < 0.01
    assert epoch_mae / 4 < 0.1



import math

import torch
import torch.nn.functional as F
from deprecated import deprecated

from torch_gauge.o3 import O3Tensor, SphericalTensor
from torch_gauge.o3.functional import NormContraction1d


class SSP(torch.nn.Softplus):
    """Shifted SoftPlus activation function."""

    def __init__(self, beta=1, threshold=20):
        super().__init__(beta, threshold)

    def forward(self, input):
        """"""
        return F.softplus(input, self.beta, self.threshold) - math.sqrt(2)


class Swish_fn(torch.autograd.Function):
    """Swish activation function."""

    @staticmethod
    def forward(ctx, i):
        """"""
        result = i * torch.sigmoid(i)
        ctx.save_for_backward(i)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        """"""
        i = ctx.saved_tensors[0]
        sigmoid_i = torch.sigmoid(i)
        return grad_output * (sigmoid_i * (1 + i * (1 - sigmoid_i)))


class Swish(torch.nn.Module):
    def forward(self, input_tensor):
        """"""
        return Swish_fn.apply(input_tensor)


class IELin(torch.nn.Module):
    r"""
    Irrep-wise Equivariant Linear Layer.

    This module takes a spherical tensor and perform linear transformation within
    the feature channels spanned by each irreducible representation index, (l, m):

    .. math::

        \mathbf{h}^{\mathrm{out}}_{l,m} = \mathbf{W}^{l} \cdot \mathbf{h}^{\mathrm{in}}_{l,m}

    The matrix multiplication is always performed on the last dimension of the spherical tensors.

    Note:
        This function is vectorized by tensor padding. An additional memory cost might
        arise, and it's recommended to used benchmark=True to minimize this overhead.

    Attributes:
        metadata_in (torch.LongTensor): the number of non-degenerate feature channels of the input tensor
            for each l (and the associated m(s)).
        metadata_out (torch.LongTensor): the number of non-degenerate feature channels of the output tensor
            for each l (and the associated m(s)).
        group (str): The group index of the tensors to be passed.
        benchmark (bool): search the optimal padding size when creating the module.
    """

    def __init__(self, metadata_in, metadata_out, group="o3", benchmark=False):
        super().__init__()
        assert metadata_in.dim() == 1
        assert len(metadata_in) == len(metadata_out)
        self._metadata_in = metadata_in
        self._metadata_out = metadata_out
        group = group.lower()
        if group == "so3":
            self.tensor_class = SphericalTensor
            self.n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0)) * 2 + 1
        elif group == "o3":
            self.tensor_class = O3Tensor
            n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0) // 2) * 2 + 1
            self.n_irreps_per_l = n_irreps_per_l.repeat_interleave(2)
        else:
            raise NotImplementedError(f"The group {group} is not supported in IELin")

        if not benchmark:
            # Maximum padding
            padding_size_in = metadata_in.max().item()
            padding_size_out = metadata_out.max().item()
        else:
            # TODO: estimate the optimal padding size and testing
            raise NotImplementedError

        # Padding rules:
        # I - non-degenerate matrices, degenerate vectors, cat reduce
        # J - non-degenerate matrices, non-degenerate vectors, sum reduce
        # m - degenerate matrices, non-degenerate vectors, cat reduce
        # Block index order: m->I->J->padding(k)
        vector_padding_idx = []
        vector_padding_mask = []
        src_offset = 0
        for lp_idx, nin_lpm in enumerate(metadata_in):
            nin_lpm = nin_lpm.item()
            nout_lpm = metadata_out[lp_idx].item()
            if metadata_in[lp_idx] == 0:
                continue
            bmv_degeneracy_lpm = -(nout_lpm // (-padding_size_out))
            if bmv_degeneracy_lpm == 0:
                src_offset += nin_lpm * self.n_irreps_per_l[lp_idx]
                continue
            for m in range(self.n_irreps_per_l[lp_idx]):
                # J-padding
                vector_padding_idx_lpm = torch.zeros(
                    -(nin_lpm // (-padding_size_in)) * padding_size_in, dtype=torch.long
                )
                vector_padding_mask_lpm = torch.zeros(
                    -(nin_lpm // (-padding_size_in)) * padding_size_in, dtype=torch.long
                )
                vector_padding_idx_lpm[:nin_lpm] = (
                    torch.arange(nin_lpm, dtype=torch.long) + src_offset
                )
                vector_padding_mask_lpm[:nin_lpm] = 1
                # I-padding
                vector_padding_idx.append(
                    vector_padding_idx_lpm.repeat(bmv_degeneracy_lpm)
                )
                vector_padding_mask.append(
                    vector_padding_mask_lpm.repeat(bmv_degeneracy_lpm)
                )
                src_offset += nin_lpm
        self.register_buffer("vector_padding_idx", torch.cat(vector_padding_idx))
        self.register_buffer(
            "vector_padding_mask", torch.cat(vector_padding_mask).bool()
        )

        matrix_select_idx = []
        matid_to_fanin = []
        mat_offset = 0
        for lp_idx, nin_lpm in enumerate(metadata_in):
            nin_lpm = nin_lpm.item()
            nout_lpm = metadata_out[lp_idx].item()
            nblocks_lp = (-(nin_lpm // (-padding_size_in))) * (
                -(nout_lpm // (-padding_size_out))
            )
            if nblocks_lp > 0:
                matrix_select_idx.append(
                    (torch.arange(nblocks_lp, dtype=torch.long) + mat_offset).repeat(
                        self.n_irreps_per_l[lp_idx]
                    )
                )
                matid_to_fanin.append(
                    torch.full((nblocks_lp,), nin_lpm, dtype=torch.long)
                )
                mat_offset += nblocks_lp
        self.register_buffer("matrix_select_idx", torch.cat(matrix_select_idx))
        self.matid_to_fanin = torch.cat(matid_to_fanin)
        self.n_mats = mat_offset
        self.n_gathered_mats = self.matrix_select_idx.shape[0]

        out_reduce_idx = []
        out_reduce_mask = []
        dst_offset = 0
        for lp_idx, nout_lpm in enumerate(metadata_out):
            nin_lpm = metadata_in[lp_idx].item()
            nout_lpm = nout_lpm.item()
            if nout_lpm == 0:
                continue
            bmv_degeneracy_lpm = -(nout_lpm // (-padding_size_out))
            in_j_degeneracy_lpm = -(nin_lpm // (-padding_size_in))
            if in_j_degeneracy_lpm == 0:
                dst_offset += nout_lpm * self.n_irreps_per_l[lp_idx]
                continue
            for m in range(self.n_irreps_per_l[lp_idx]):
                dst_idx_lpm = (
                    (
                        dst_offset
                        + torch.arange(
                            padding_size_out * bmv_degeneracy_lpm, dtype=torch.long
                        )
                    )
                    .view(bmv_degeneracy_lpm, 1, padding_size_out)
                    .expand(bmv_degeneracy_lpm, in_j_degeneracy_lpm, padding_size_out)
                    .contiguous()
                    .view(-1)
                )
                dst_mask_lpm = torch.zeros(
                    padding_size_out * bmv_degeneracy_lpm, dtype=torch.long
                )
                dst_mask_lpm[:nout_lpm] = 1
                dst_mask_lpm = (
                    dst_mask_lpm.view(bmv_degeneracy_lpm, 1, padding_size_out)
                    .expand(bmv_degeneracy_lpm, in_j_degeneracy_lpm, padding_size_out)
                    .contiguous()
                    .view(-1)
                )
                out_reduce_idx.append(dst_idx_lpm)
                out_reduce_mask.append(dst_mask_lpm)
                dst_offset += nout_lpm
        self.register_buffer("out_reduce_mask", torch.cat(out_reduce_mask).bool())
        self.register_buffer(
            "out_reduce_idx",
            torch.cat(out_reduce_idx)[self.out_reduce_mask],
        )

        self.linears = torch.nn.Parameter(
            torch.zeros(self.n_mats, padding_size_in, padding_size_out)
        )
        self.padding_size_in = padding_size_in
        self.padding_size_out = padding_size_out

        self.register_buffer(
            "out_layout",
            self.tensor_class.generate_rep_layout_1d_(self._metadata_out),
        )
        self.num_out_channels = torch.sum(self._metadata_out).item()
        self.reset_parameters()

    def reset_parameters(self) -> None:
        with torch.no_grad():
            for lid, fan_in in enumerate(self.matid_to_fanin):
                bound = torch.sqrt(1 / fan_in)
                self.linears.data[lid].uniform_(-bound, bound)

    def forward(self, x: SphericalTensor) -> SphericalTensor:
        """
        Args:
            x: The input SphericalTensor.

        Returns:
            The data tensor of the output spherical tensor.
        """
        assert len(x.rep_dims) == 1
        assert x.rep_dims[-1] == x.ten.dim() - 1
        assert torch.all(x.metadata[-1].eq(self._metadata_in)), (
            f"Expected the SphericalTensor x and self._metadata_in to have the "
            f"same irrep metadata along the last dimension, got {x.metadata[-1]}"
            f" and {self._metadata_in} instead"
        )
        in_ten = x.ten.view(-1, x.ten.shape[-1])
        padded_in_ten = torch.index_select(
            in_ten, dim=1, index=self.vector_padding_idx
        ).mul_(self.vector_padding_mask)
        # to (lIJ)-B-k layout
        padded_in_ten = padded_in_ten.view(
            in_ten.shape[0], self.n_gathered_mats, self.padding_size_in
        ).transpose(0, 1)
        gathered_linears = torch.index_select(
            self.linears, dim=0, index=self.matrix_select_idx
        )
        padded_out_ten = (
            torch.bmm(padded_in_ten, gathered_linears)
            .transpose(0, 1)
            .contiguous()
            .view(in_ten.shape[0], -1)
        )
        out_ten = torch.zeros(
            in_ten.shape[0],
            self.out_layout.shape[1],
            dtype=in_ten.dtype,
            device=in_ten.device,
        ).index_add_(1, self.out_reduce_idx, padded_out_ten[:, self.out_reduce_mask])
        out_ten = out_ten.view(*x.ten.shape[:-1], -1)

        # out_metadata = x.metadata.clone()
        # out_metadata[-1] = self._metadata_out
        out_metadata = self._metadata_out.unsqueeze(0)
        return self.tensor_class(
            out_ten,
            rep_dims=x.rep_dims,
            metadata=out_metadata,
            rep_layout=x.rep_layout[:-1] + (self.out_layout.data,),
            num_channels=x.num_channels[:-1] + (self.num_out_channels,),
        )


class IELinSerial(torch.nn.Module):
    r"""
    Irrep-wise Equivariant Linear Layer.

    This module takes a spherical tensor and perform linear transformation within
    the feature channels spanned by each irreducible representation index, (l, m):

    .. math::

        \mathbf{h}^{\mathrm{out}}_{l,m} = \mathbf{W}^{l} \cdot \mathbf{h}^{\mathrm{in}}_{l,m}

    The matrix multiplication is always performed on the last dimension of the spherical tensors.

    Note:
        The implementation of this operation is not optimized.

    Attributes:
        metadata_in (torch.LongTensor): the number of non-degenerate feature channels of the input tensor
            for each l (and the associated m(s)).
        metadata_out (torch.LongTensor): the number of non-degenerate feature channels of the output tensor
            for each l (and the associated m(s)).
        group (str): The group index of the tensors to be passed.
    """

    @deprecated(version="1.4.0", reason="Please use the vectorized LELin module")
    def __init__(self, metadata_in, metadata_out, group="o3"):
        super().__init__()
        assert metadata_in.dim() == 1
        assert len(metadata_in) == len(metadata_out)
        self._metadata_in = metadata_in
        self._metadata_out = metadata_out
        group = group.lower()
        if group == "so3":
            self.tensor_class = SphericalTensor
            self.n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0)) * 2 + 1
        elif group == "o3":
            self.tensor_class = O3Tensor
            n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0) // 2) * 2 + 1
            self.n_irreps_per_l = n_irreps_per_l.repeat_interleave(2)
        else:
            raise NotImplementedError(f"The group {group} is not supported in IELin")
        # The bias must be False
        self.linears = torch.nn.ModuleList(
            [
                torch.nn.Linear(
                    metadata_in[l],
                    metadata_out[l],
                    bias=False,
                )
                if self._metadata_out[l] > 0 and self._metadata_in[l] > 0
                else None
                for l, _ in enumerate(metadata_in)
            ]
        )
        self._end_inds_in = torch.cumsum(self._metadata_in * self.n_irreps_per_l, dim=0)
        self._start_inds_in = torch.cat([torch.LongTensor([0]), self._end_inds_in[:-1]])
        self.register_buffer(
            "out_layout",
            self.tensor_class.generate_rep_layout_1d_(self._metadata_out),
        )
        self.num_out_channels = torch.sum(self._metadata_out).item()
        self.reset_parameters()

    def reset_parameters(self) -> None:
        for linear in self.linears:
            if linear is None:
                continue
            torch.nn.init.kaiming_uniform_(linear.weight, a=math.sqrt(5), mode="fan_in")

    def forward(self, x: SphericalTensor) -> SphericalTensor:
        """
        Args:
            x: The input SphericalTensor.

        Returns:
            The data tensor of the output spherical tensor.
        """
        assert x.rep_dims[-1] == x.ten.dim() - 1
        assert torch.all(x.metadata[-1].eq(self._metadata_in)), (
            f"Expected the SphericalTensor x and self._metadata_in to have the "
            f"same irrep metadata along the last dimension, got {x.metadata[-1]}"
            f" and {self._metadata_in} instead"
        )
        outs = []
        for l, linear_l in enumerate(self.linears):
            if linear_l is None:
                if self._metadata_out[l] > 0:
                    outs.append(
                        torch.zeros(
                            *x.shape[:-1],
                            self._metadata_out[l] * self.n_irreps_per_l[l],
                            dtype=x.ten.dtype,
                            device=x.ten.device,
                        )
                    )
                continue
            in_l = x.ten[
                ..., self._start_inds_in[l] : self._end_inds_in[l]
            ].contiguous()
            out_l = linear_l(
                in_l.unflatten(-1, (self.n_irreps_per_l[l], self._metadata_in[l]))
            )
            outs.append(
                out_l.view(
                    *x.shape[:-1], self._metadata_out[l] * self.n_irreps_per_l[l]
                )
            )

        out_ten = torch.cat(outs, dim=-1)
        out_metadata = x.metadata.clone()
        out_metadata[-1] = self._metadata_out
        out_rep_layout = x.rep_layout[:-1] + (self.out_layout.data,)
        return self.tensor_class(
            out_ten,
            rep_dims=x.rep_dims,
            metadata=out_metadata,
            rep_layout=out_rep_layout,
            num_channels=x.num_channels[:-1] + (self.num_out_channels,),
        )


class RepNorm1d(torch.nn.Module):
    r"""
    The (experimental) Representation Normalization layer.

    .. math::

        \mathrm{RepNorm}(\mathbf{h})_{l,m} = \mathrm{Norm}(||\mathbf{h}_l||)
        \cdot \frac{\mathbf{h}_{l}}{ ||\mathbf{h}_{l}|| + \beta + \epsilon}

    Heuristically, the trainable :math:`\mathbf{\beta}` controls the fraction of norm
    information to be retained.

    The specified invariant channels are not affected by the gauge factorization, i.e.
    with the pure gauge regarded as ones.
    """

    def __init__(
        self,
        num_channels,
        norm="batch",
        momentum=0.1,
        eps=1e-2,
        n_invariant_channels=0,
        mode="raw",
    ):
        super().__init__()
        self._num_channels = num_channels
        self._n_invariant_channels = n_invariant_channels
        self._eps = eps
        self._mode = mode
        if norm == "batch":
            self.norm = torch.nn.BatchNorm1d(
                num_features=num_channels, momentum=momentum, affine=False
            )
        elif norm == "node":
            self.norm = torch.nn.LayerNorm(
                normalized_shape=num_channels, elementwise_affine=False
            )
        elif norm == "none":
            self.norm = None
        else:
            raise NotImplementedError
        if self._mode == "raw":
            self.beta = torch.nn.Parameter(
                0.5 + torch.rand(self._num_channels - self._n_invariant_channels) * 1.0
            )
        elif self._mode == "inv":
            self.betainv = torch.nn.Parameter(
                0.5 + torch.rand(self._num_channels - self._n_invariant_channels) * 1.5
            )
        else:
            raise NotImplementedError

    def forward(self, x: SphericalTensor) -> (torch.Tensor, SphericalTensor):
        """
        Args:
            x: The input SphericalTensor.

        Returns:
            (tuple): tuple containing:
                (torch.Tensor): The normalized invariant content.

                (SphericalTensor): The "pure gauge" spherical tensor.
        """
        if self.norm is None:
            x0 = NormContraction1d.apply(
                x.ten, x.rep_layout[0][2], (x.ten.shape[0], x.num_channels[0]), 1, 1e-04
            )
            return x0, x
        assert x.ten.dim() == 2
        if self._mode == "raw":
            beta = self.beta
        elif self._mode == "inv":
            beta = self.betainv.abs().add(self._eps).reciprocal()
        else:
            raise NotImplementedError
        if self._n_invariant_channels == 0:
            x0 = NormContraction1d.apply(
                x.ten, x.rep_layout[0][2], (x.ten.shape[0], x.num_channels[0]), 1, 1e-04
            )
            x1 = self.norm(x0)
            divisor = x0.add(beta).abs().add(self._eps)
            divisor_broadcasted = torch.index_select(
                divisor,
                dim=1,
                index=x.rep_layout[0][2],
            )
            x2_ten = x.ten.div(divisor_broadcasted)
        else:
            x0 = x.invariant()
            assert self._n_invariant_channels <= x.metadata[0][0]
            x1 = self.norm(x0)
            divisor = (
                torch.abs(x0[:, self._n_invariant_channels :].add(beta)) + self._eps
            )
            divisor_broadcasted = torch.index_select(
                divisor,
                dim=1,
                index=(
                    x.rep_layout[0][2, self._n_invariant_channels :]
                    - self._n_invariant_channels
                ),
            )
            x2_ten = torch.cat(
                [
                    torch.ones_like(x1[:, : self._n_invariant_channels]),
                    x.ten[:, self._n_invariant_channels :].div(divisor_broadcasted),
                ],
                dim=1,
            )
        x2 = x.self_like(x2_ten)
        # x2.ten = x2.ten.div((x.rep_layout[0][0] * 2 + 1).type(x0.dtype).sqrt().unsqueeze(0))
        return x1, x2


class KernelBroadcast(torch.nn.Module):
    """
    Specialized module for broadcasting scalar features with
    a SO(3) kernel of singleton channel-sizes per angular quantum number l.
    """

    def __init__(self, target_metadata):
        super().__init__()
        self.in_num_channels = len(target_metadata)
        self.target_num_channels = torch.sum(target_metadata).item()
        filter_broadcast_idx = []
        src_idx = 0
        for l, n_lm in enumerate(target_metadata):
            for m in range(2 * l + 1):
                filter_broadcast_idx.append(
                    torch.full((n_lm.item(),), src_idx, dtype=torch.long)
                )
                src_idx += 1

        self.register_buffer("filter_broadcast_idx", torch.cat(filter_broadcast_idx))
        self.register_buffer(
            "feat_broadcast_idx",
            SphericalTensor.generate_rep_layout_1d_(target_metadata)[2, :],
        )

    def forward(self, rshs: SphericalTensor, feat: torch.Tensor) -> torch.Tensor:
        assert self.in_num_channels == rshs.num_channels[0]
        assert self.target_num_channels == feat.shape[-1]
        broadcasted_rshs = torch.index_select(
            rshs.ten, dim=rshs.ten.dim() - 1, index=self.filter_broadcast_idx
        )
        broadcasted_feat = torch.index_select(
            feat, dim=feat.dim() - 1, index=self.feat_broadcast_idx
        )
        return broadcasted_feat * broadcasted_rshs



import os
from torch_gauge.verlet_list import VerletList
from torch_gauge._version import get_versions

__version__ = get_versions()["version"]
del get_versions

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))



import torch

from torch_gauge.o3 import O3Tensor, SphericalTensor


def test_spherical_tensor_creation1d():
    dten = torch.rand(4, 6, 12, 101, 7)
    metadata = torch.LongTensor([[7, 9, 1, 5, 3]])
    SphericalTensor(dten, (3,), metadata)
    return 0


def test_spherical_tensor_creation2d():
    dten = torch.rand(4, 6, 12, 101, 7)
    metadata = torch.LongTensor([[1, 2, 1, 0, 0], [7, 9, 1, 5, 3]])
    SphericalTensor(dten, (2, 3), metadata)
    return 0


def test_spherical_tensor_layout1d():
    dten = torch.rand(4, 19)
    metadata = torch.LongTensor([[1, 2, 1, 1]])
    test_sp_ten = SphericalTensor(dten, (1,), metadata)
    assert test_sp_ten.num_channels == (5,)
    assert torch.all(
        test_sp_ten.rep_layout[0][0].eq(
            torch.LongTensor([0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3])
        )
    )
    assert torch.all(
        test_sp_ten.rep_layout[0][1].eq(
            torch.LongTensor(
                [0, -1, -1, 0, 0, 1, 1, -2, -1, 0, 1, 2, -3, -2, -1, 0, 1, 2, 3]
            )
        )
    )
    assert torch.all(
        test_sp_ten.rep_layout[0][2].eq(
            torch.LongTensor([0, 1, 2, 1, 2, 1, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4])
        )
    )
    return 0


def test_spherical_tensor_scalar_product_1d():
    dten = torch.tensor(
        [
            [1.1, 2.2, 0.5, 0.6, -0.6, 0.9, 0.7, 0.3, 0.1, 0.2, 0.3, 0.4, 0.5],
            [0.0, 0.0, 0.2, 0.0, 0.2, 0.0, 0.3, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        ]
    )
    metadata = torch.LongTensor([[2, 2, 1]])
    test_sp_ten = SphericalTensor(dten, (1,), metadata)
    scale_ten = torch.tensor([[2.0, 0.5, 3.0, 4.0, 1.0], [101.0, 7.0, 8.0, 82.0, 4.0]])
    dten_outplace = test_sp_ten.scalar_mul(scale_ten)
    assert torch.allclose(
        dten_outplace.ten,
        torch.tensor(
            [
                [2.2, 1.1, 1.5, 2.4, -1.8, 3.6, 2.1, 1.2, 0.1, 0.2, 0.3, 0.4, 0.5],
                [0.0, 0.0, 1.6, 0.0, 1.6, 0.0, 2.4, 0.0, 4.0, 4.0, 4.0, 4.0, 4.0],
            ]
        ),
        atol=1e-10,
        rtol=1e-7,
    )
    assert torch.all(dten_outplace.rep_layout[0].eq(test_sp_ten.rep_layout[0]))
    dten_inplace = test_sp_ten.scalar_mul(scale_ten, inplace=True)
    assert torch.allclose(
        dten_inplace.ten,
        torch.tensor(
            [
                [2.2, 1.1, 1.5, 2.4, -1.8, 3.6, 2.1, 1.2, 0.1, 0.2, 0.3, 0.4, 0.5],
                [0.0, 0.0, 1.6, 0.0, 1.6, 0.0, 2.4, 0.0, 4.0, 4.0, 4.0, 4.0, 4.0],
            ]
        ),
        atol=1e-10,
        rtol=1e-7,
    )
    assert torch.all(dten_inplace.rep_layout[0].eq(test_sp_ten.rep_layout[0]))


def test_spherical_tensor_scalar_product_2d():
    metadata2d = torch.LongTensor([[2, 1], [1, 1]])
    test_2d = SphericalTensor(
        torch.tensor(
            [
                [0.1, 0.3, 0.3, 0.3],
                [0.2, -0.7, 0.1, -0.2],
                [0.3, 0.2, -0.3, 0.4],
                [0.2, 0.3, 0.4, -0.5],
                [0.1, 0.2, 0.3, 0.4],
            ]
        ),
        (
            0,
            1,
        ),
        metadata2d,
    )
    scale_2d = torch.tensor(
        [
            [1.0, 2.0],
            [3.0, 4.0],
            [5.0, 6.0],
        ]
    )
    out_ten = test_2d.scalar_mul(scale_2d).ten
    ref_ten = torch.tensor(
        [
            [0.1, 0.6, 0.6, 0.6],
            [0.6, -2.8, 0.4, -0.8],
            [1.5, 1.2, -1.8, 2.4],
            [1.0, 1.8, 2.4, -3.0],
            [0.5, 1.2, 1.8, 2.4],
        ]
    )
    assert torch.allclose(out_ten, ref_ten, atol=1e-7, rtol=1e-5)


def test_spherical_tensor_dot_product():
    dten1 = torch.rand(4, 6, 1, 101, 7)
    metadata = torch.LongTensor([[7, 9, 1, 5, 3]])
    test1 = SphericalTensor(dten1, (3,), metadata)
    dten2 = torch.rand(4, 6, 12, 101, 7)
    metadata = torch.LongTensor([[1, 2, 1, 0, 0], [7, 9, 1, 5, 3]])
    test2 = SphericalTensor(dten2, (2, 3), metadata)
    test_dot = test2.dot(test1, dim=3)
    assert torch.all(test_dot.ten.eq(test1.ten.mul(test2.ten).sum(3)))
    assert test_dot.shape == (4, 6, 12, 7)
    assert test_dot.rep_dims == (2,)
    assert torch.all(test_dot.metadata.eq(torch.LongTensor([[1, 2, 1, 0, 0]])))
    assert torch.all(test_dot.rep_layout[0].eq(test2.rep_layout[0]))


def test_spherical_tensor_rep_dot():
    # Minimal 1d example
    metadata = torch.LongTensor([[1, 2]])
    test_1d_1 = SphericalTensor(
        torch.tensor([0.2, 0.1, 0.2, 0.3, 0.2, 0.3, 0.4]), (0,), metadata
    )
    test_1d_2 = SphericalTensor(
        torch.tensor([0.4, 0.2, 0.3, -0.4, -0.4, -0.3, 0.4]), (0,), metadata
    )
    test_1d_out = test_1d_1.rep_dot(test_1d_2, dim=0)
    assert torch.allclose(
        test_1d_out,
        torch.tensor([0.08, -0.19, 0.14]),
        atol=1e-9,
        rtol=1e-7,
    )

    # 2d shape test
    dten1 = torch.rand(4, 6, 1, 101, 7)
    metadata = torch.LongTensor([[7, 9, 1, 5, 3]])
    test1 = SphericalTensor(dten1, (3,), metadata)
    dten2 = torch.rand(4, 6, 12, 101, 7)
    metadata = torch.LongTensor([[1, 2, 1, 0, 0], [7, 9, 1, 5, 3]])
    test2 = SphericalTensor(dten2, (2, 3), metadata)
    test_dot = test2.rep_dot(test1, dim=3)
    assert test_dot.shape == (4, 6, 12, 25, 7)
    assert test_dot.rep_dims == (2,)
    assert torch.all(test_dot.metadata.eq(torch.LongTensor([[1, 2, 1, 0, 0]])))
    assert torch.all(test_dot.rep_layout[0].eq(test2.rep_layout[0]))

    # When L=0 entries are positive, self rep-dot should return a
    # tensor almost the same as the invariant content within threshold _norm_eps
    test3 = SphericalTensor(
        torch.rand(32, 11, 5, 101, 7), (3,), torch.LongTensor([[7, 9, 1, 5, 3]])
    )
    dot_3 = test3.rep_dot(test3, dim=3)
    assert torch.allclose(
        dot_3,
        test3.invariant().pow(2),
        atol=1e-4,
        rtol=1e-3,
    )


def test_spherical_tensor_rep_outer():
    dten1 = torch.rand(101)
    metadata = torch.LongTensor([[7, 9, 1, 5, 3]])
    test1 = SphericalTensor(dten1, (0,), metadata)
    dten2 = torch.rand(101)
    test2 = SphericalTensor(dten2, (0,), metadata)
    test_outer = test1.rep_outer(test2)
    assert test_outer.shape == (101, 101)
    assert torch.all(test_outer.ten.eq(torch.outer(dten1, dten2)))
    assert torch.all(
        test_outer.metadata.eq(torch.tensor([[7, 9, 1, 5, 3], [7, 9, 1, 5, 3]]))
    )
    test3 = SphericalTensor(torch.rand(4, 6, 12, 101, 7), (3,), metadata)
    test4 = SphericalTensor(
        torch.rand(4, 6, 12, 12, 7), (3,), torch.LongTensor([[1, 2, 1, 0, 0]])
    )
    test_outer = test3.rep_outer(test4)
    assert test_outer.shape == (4, 6, 12, 101, 12, 7)
    assert test_outer.rep_dims == (3, 4)
    assert torch.all(
        test_outer.metadata.eq(torch.tensor([[7, 9, 1, 5, 3], [1, 2, 1, 0, 0]]))
    )


def test_spherical_tensor_invariant_contraction():
    metadata1d = torch.LongTensor([[2, 2]])
    test_1d = SphericalTensor(
        torch.tensor([0.2, -0.7, 0.1, -0.2, 0.3, 0.2, -0.3, 0.4]), (0,), metadata1d
    )
    assert torch.allclose(
        test_1d.invariant(),
        torch.tensor([0.2, -0.7, 0.4358, 0.4898]),
        atol=1e-4,
        rtol=0,
    )

    metadata2d = torch.LongTensor([[1, 1], [1, 1]])
    test_2d = SphericalTensor(
        torch.tensor(
            [
                [0.2, -0.7, 0.1, -0.2],
                [0.3, 0.2, -0.3, 0.4],
                [0.2, 0.3, 0.4, -0.5],
                [0.1, 0.2, 0.3, 0.4],
            ]
        ),
        (
            0,
            1,
        ),
        metadata2d,
    )
    assert torch.allclose(
        test_2d.invariant(),
        torch.tensor([[0.2, 0.73484], [0.37416, 1.03923]]),
        atol=1e-4,
        rtol=0,
    )


def test_spherical_tensor_fold_dim():
    metadata = torch.LongTensor([[8, 4, 2, 4, 0]])
    test = SphericalTensor(torch.rand(4, 6, 12, 58, 7), (3,), metadata)
    folded = test.fold(stride=2)
    assert folded.shape == (4, 6, 12, 29, 2, 7)
    assert torch.all(folded.metadata[0].eq(torch.LongTensor([4, 2, 1, 2, 0])))
    assert folded.num_channels == (9,)

    # Test error capturing
    metadata = torch.LongTensor([[8, 4, 2, 3, 1]])
    test = SphericalTensor(torch.rand(4, 6, 12, 60, 7), (3,), metadata)
    try:
        test.fold(stride=2)
    except AssertionError as e:
        assert (
            str(e)
            == f"The number of channels for the SphericalTensor to be folded must be multiples of "
            "stride, got (tensor([[8, 4, 2, 3, 1]]), 2) instead"
        )


def test_to_o3():
    dten1 = torch.rand(4, 6, 1, 101, 7)
    metadata = torch.LongTensor([[7, 9, 1, 5, 3]])
    test1 = SphericalTensor(dten1, (3,), metadata)
    o3_ten1 = O3Tensor.from_so3(test1, parity=1)
    assert o3_ten1.rep_dims == (3,)
    assert torch.all(
        o3_ten1.metadata.eq(torch.LongTensor([[7, 0, 9, 0, 1, 0, 5, 0, 3, 0]]))
    )
    assert torch.all(o3_ten1.rep_layout[0][3].eq(torch.ones(101, dtype=torch.long)))
    assert torch.all(o3_ten1.rep_layout[0][:3].eq(test1.rep_layout[0]))
    o3_ten2 = O3Tensor.from_so3(test1, parity=-1)
    assert o3_ten2.rep_dims == (3,)
    assert torch.all(
        o3_ten2.metadata.eq(torch.LongTensor([[0, 7, 0, 9, 0, 1, 0, 5, 0, 3]]))
    )
    assert torch.all(o3_ten2.rep_layout[0][3].eq(-torch.ones(101, dtype=torch.long)))
    assert torch.all(o3_ten2.rep_layout[0][:3].eq(test1.rep_layout[0]))
